from typing import List, Any, Dict, Optional
import httpx
import litellm
from litellm import acompletion, completion


def make_request(
    model_name: str,
    max_retries: int,
    api_key: str,
    messages: List[Dict[str, Any]],
    tools: List[Dict[str, Any]],
    tool_choice: Dict[str, Any],
    organization: Optional[str],
    api_base: Optional[str]
):
    if len(messages) == 0:
        raise ValueError("No messages provided")

    req = {
        "model": model_name,
        "messages": messages,
        "max_retries": max_retries,
        "api_key": api_key,
    }

    if len(tools) > 0:
        req["tools"] = tools
        req["tool_choice"] = tool_choice

    if organization:
        req["organization"] = organization
    if api_base:
        req["api_base"] = api_base

    return req


def pre_request_setup(
        ssl_verify: bool,
        verbose: bool,
        headers: Optional[Dict[str, str]]
):
    if not ssl_verify:
        litellm.client_session = httpx.Client(verify=False)

    if verbose:
        litellm.set_verbose = True

    if headers is not None:
        litellm.headers = headers


def post_request_cleanup(
    ssl_verify: bool,
    verbose: bool,
    headers: Optional[Dict[str, str]]
):
    if not ssl_verify:
        litellm.client_session = None  # reset to default. Might be unnecessary. Evaluate later.

    if headers is not None:
        litellm.headers = None  # reset to default. Might be unnecessary. Evaluate later.

    if verbose:
        litellm.set_verbose = False


def litellm_request(
        ssl_verify: bool,
        model_name: str,
        max_retries: int,
        api_key: str,
        messages: List[Dict[str, Any]],
        tools: List[Dict[str, Any]],
        tool_choice: Dict[str, Any],
        verbose: bool = False,
        headers: Optional[Dict[str, str]] = None,
        organization: Optional[str] = None,
        api_base: Optional[str] = None
) -> Any:
    req = make_request(
        model_name,
        max_retries,
        api_key,
        messages,
        tools,
        tool_choice,
        organization,
        api_base
    )
    pre_request_setup(ssl_verify, verbose, headers)
    response = completion(**req)
    post_request_cleanup(ssl_verify, verbose, headers)
    return response


async def async_litellm_request(
        ssl_verify: bool,
        model_name: str,
        max_retries: int,
        api_key: str,
        messages: List[Dict[str, Any]],
        tools: List[Dict[str, Any]],
        tool_choice: Dict[str, Any],
        verbose: bool = False,
        headers: Optional[Dict[str, str]] = None,
        organization: Optional[str] = None,
        api_base: Optional[str] = None
) -> Any:
    req = make_request(
        model_name,
        max_retries,
        api_key,
        messages,
        tools,
        tool_choice,
        organization,
        api_base
    )
    pre_request_setup(ssl_verify, verbose, headers)
    response = await acompletion(**req)
    post_request_cleanup(ssl_verify, verbose, headers)
    return response



from typing import ClassVar, Optional, Type
from typedtemplate import TypedTemplate, BaseTemplateEngine
from pydantic import BaseModel


class TypedPrompt(BaseModel):
    """
    This is a placeholder class to implement automated prompt engineering with DSPy later.
    """
    static_instructions: ClassVar[Optional[str]] = None
    instructions: Optional[TypedTemplate] = None
    context: Optional[TypedTemplate] = None

    def render(self) -> str:
        if self.instructions is None and self.static_instructions is None:
            raise ValueError("No instructions provided. Both static_instructions and instructions are None. "
                             "One must be provided.")
        ins = self.instructions.render() if self.instructions is not None else self.static_instructions
        ctx = f"\n# Context\n{self.context.render()}" if self.context is not None else ""
        return f"# Instructions \n{ins}{ctx}"




"""The `version` module holds the version information for TypedLLM."""
from __future__ import annotations as _annotations

__all__ = 'VERSION', 'version_short'

VERSION = '0.3.1'
"""The version of TypedLLM."""


def version_short() -> str:
    """Return the `major.minor` part of TypedLLM version.

    It returns '2.1' if TypedLLM version is '2.1.1'.
    """
    return '.'.join(VERSION.split('.')[:2])



import inspect
from typing import Type, Callable, List
from pydantic import BaseModel, Field, create_model


class Tool(BaseModel):
    """LLM Tool that can be used in a request"""
    def __call__(self) -> any:
        """
        Implement this function to make the tool callable.
        :return:
        """
        raise NotImplementedError("The __call__ method is not implemented for the Tool class.")

    @classmethod
    def openai_tool_choice_json(cls):
        return {
            "type": "function",
            "function": {
                "name": cls.__name__,
            }
        }

    @classmethod
    def openapi_json(cls):
        name = cls.__name__
        description = cls.__doc__ or f"Tool called {cls.__name__}"
        if cls.__doc__ is None:
            # Patch docstring so Tool classes and functions act the same. See unit tests.
            cls.__doc__ = description
        schema = cls.model_json_schema()
        result = {
            "type": "function",
            "function": {
                "name": name,
                "description": description,
                "parameters": schema
            }
        }
        return result


StringClass = "a".__class__


def create_tool_from_function(
        function: Callable,
) -> Type[Tool]:
    sig = inspect.signature(function)
    annotations = {}
    for name, param in sig.parameters.items():
        if param.annotation is param.empty:
            raise ValueError(f"Parameter {name} does not have a type hint")
        annotations[name] = (param.annotation, Field(...))
    ToolClz: Type[Tool] = create_model(
        function.__name__,
        **annotations,
        __base__=(Tool,),
        __doc__=function.__doc__ or f"Tool called {function.__name__}"
    )
    return ToolClz


class ToolCollection(List[Type[Tool]]):
    def __init__(self, *args):
        super().__init__(args)

    def openapi_json(self):
        return [tool.openapi_json() for tool in self]

    def get_by_name(self, name: str) -> Type[Tool]:
        for tool in self:
            if tool.__name__ == name:
                return tool
        raise ValueError(f"Tool with name {name} not found.")



import json
from typing import Type, List, Optional

import pydantic
from typedtemplate import TypedTemplate

from .interop import litellm_request, async_litellm_request
from .tool import ToolCollection, Tool
from .models import (
    LLMSession,
    LLMRequest,
    LLMResponse,
    LLMAssistantMessage,
    LLMAssistantToolCall, LLMModel, LLMUserMessage
)


async def async_request(
        session: LLMSession,
        req: LLMRequest
) -> (LLMSession, LLMResponse):
    # TODO: Add support for a Memory Interface here.
    # Memory can be set as a configuration option and can be invoked here.
    messages = generate_message_json(
        *session.messages,
        *req.tool_results,
        req.message
    )
    tools, tool_choice = generate_tool_json(
        *session.tools, *req.tools,
        force_text_response=req.force_text_response,
        required_tool=req.required_tool
    )
    raw_response = await async_litellm_request(
        session.model.ssl_verify,
        session.model.name,
        session.model.max_retries,
        session.model.api_key,
        messages,
        tools.openapi_json(),
        tool_choice,
        verbose=session.verbose,
        headers=session.model.headers,
        organization=session.model.organization,
        api_base=session.model.api_base
    )

    # This is broken out to support sync/async code paths in the tool calls section
    assistant_message = extract_assistant_message(raw_response)
    tools_calls = await async_extract_tool_calls(raw_response, tools, session)
    response = LLMResponse(
        message=assistant_message,
        tool_calls=tools_calls,
        raw=raw_response
    )

    update_session(session, response)
    return session, response


def request(
        session: LLMSession,
        req: LLMRequest
) -> (LLMSession, LLMResponse):
    # TODO: Add support for a Memory Interface here.
    # Memory can be set as a configuration option and can be invoked here.
    messages = generate_message_json(
        *session.messages,
        *req.tool_results,
        req.message
    )
    tools, tool_choice = generate_tool_json(
        *session.tools, *req.tools,
        force_text_response=req.force_text_response,
        required_tool=req.required_tool
    )
    raw_response = litellm_request(
        session.model.ssl_verify,
        session.model.name,
        session.model.max_retries,
        session.model.api_key,
        messages,
        tools.openapi_json(),
        tool_choice,
        verbose=session.verbose,
        headers=session.model.headers,
        organization=session.model.organization,
        api_base=session.model.api_base
    )

    # This is broken out to support sync/async code paths in the tool calls section
    assistant_message = extract_assistant_message(raw_response)
    tools_calls = sync_extract_tool_calls(raw_response, tools, session)
    response = LLMResponse(
        message=assistant_message,
        tool_calls=tools_calls,
        raw=raw_response
    )

    update_session(session, response)
    return session, response


def update_session(session: LLMSession, response: LLMResponse) -> None:
    # Add tool calls to session messsages
    if response.tool_calls and len(response.tool_calls) > 0:
        for tool_call in response.tool_calls:
            session.messages.append(tool_call)

    # Add messages to the session
    if response.message:
        session.messages.append(response.message)


def generate_message_json(*messages):
    return [msg.openai_json() for msg in messages]


def generate_tool_json(*tools, force_text_response=False, required_tool=None):
    tools = ToolCollection(*tools)
    if len(tools) == 0 or force_text_response:
        tool_choice = "none"
    elif required_tool:
        tool_choice = required_tool.openai_tool_choice_json()
    else:
        tool_choice = "auto"
    return tools, tool_choice


def extract_assistant_message(res) -> Optional[LLMAssistantMessage]:
    from litellm.utils import ModelResponse
    response: ModelResponse = res

    if len(response.choices) != 1:
        raise Exception("Invalid number of choices in response. Expect only one choice.")

    msg = response.choices[0].message

    if msg["role"] != "assistant":
        raise Exception("Invalid role in response")
    if msg.content:
        return LLMAssistantMessage(
            content=msg.content,
        )
    return None


async def async_extract_tool_calls(
        res,  # LiteLLM Response
        tools: ToolCollection,
        session: LLMSession,
        max_depth: int = 3
) -> List[LLMAssistantToolCall]:
    msg = res.choices[0].message
    tool_calls = []
    if hasattr(msg, "tool_calls") and len(msg.tool_calls) > 0:
        for tool_call in msg["tool_calls"]:
            ToolClz = tools.get_by_name(tool_call.function.name)
            tool = await async_tool_from_json_str(
                session,
                tool_call.function.arguments,
                ToolClz,
                max_depth=max_depth
            )
            tc = LLMAssistantToolCall(
                id=tool_call.id,
                tool=tool,
                args=tool_call.function.arguments
            )
            tool_calls.append(tc)
    return tool_calls


def sync_extract_tool_calls(
        res,
        tools: ToolCollection,
        session: LLMSession,
        max_depth: int = 3
) -> List[LLMAssistantToolCall]:
    msg = res.choices[0].message
    tool_calls = []
    if hasattr(msg, "tool_calls") and len(msg.tool_calls) > 0:
        for tool_call in msg["tool_calls"]:
            ToolClz = tools.get_by_name(tool_call.function.name)
            tool = sync_tool_from_json_str(
                session,
                tool_call.function.arguments,
                ToolClz,
                max_depth=max_depth
            )
            tc = LLMAssistantToolCall(
                id=tool_call.id,
                tool=tool,
                args=tool_call.function.arguments
            )
            tool_calls.append(tc)
    return tool_calls


def typed_request(model: LLMModel, template: Type[TypedTemplate], tool: Type[Tool], **kwargs):
    msg = template(**kwargs).render()
    session = LLMSession(model=model)
    req = LLMRequest(
        tools=[tool],
        required_tool=tool,
        message=LLMUserMessage(content=msg)
    )
    _, response = request(session, req)
    tool_response = response.tool_calls[0].tool
    return tool_response


async def async_typed_request(model: LLMModel, template: Type[TypedTemplate], tool: Type[Tool], **kwargs):
    msg = template(**kwargs).render()
    session = LLMSession(model=model)
    req = LLMRequest(
        tools=[tool],
        required_tool=tool,
        message=LLMUserMessage(content=msg)
    )
    _, response = await async_request(session, req)
    tool_response = response.tool_calls[0].tool
    return tool_response


async def async_tool_from_json_str(session: LLMSession, json_str: str, ToolClz: Type[Tool], depth: int = 0, max_depth: int = 3) -> Tool:
    try:
        argument_dict = json.loads(json_str)
        tool = ToolClz(**argument_dict)
        return tool
    except (json.JSONDecodeError, pydantic.ValidationError) as e:
        if depth >= max_depth:
            raise Exception("Invalid Tool JSON. Max depth reached.")

        msg = f"""# Instructions
        You must fix the invalid JSON provided to match the tool.
        Use the error message below to fix the JSON.
        
        # JSON
        {json_str}
        
        # Error
        {str(e)}
        """
        tools, tool_choice = generate_tool_json(
            ToolClz,
            force_text_response=False,
            required_tool=ToolClz
        )
        response = await async_litellm_request(
            session.model.ssl_verify,
            session.model.name,
            session.model.max_retries,
            session.model.api_key,
            generate_message_json(LLMUserMessage(content=msg)),
            tools.openapi_json(),
            tool_choice,
            verbose=session.verbose,
            headers=session.model.headers,
            organization=session.model.organization,
            api_base=session.model.api_base
        )
        return await async_tool_from_json_str(
            session,
            response.choices[0].message.tool_calls[0].function.arguments,
            ToolClz,
            depth + 1,
            max_depth
        )


def sync_tool_from_json_str(session: LLMSession, json_str: str, ToolClz: Type[Tool], depth: int = 0, max_depth: int = 3) -> Tool:
    try:
        argument_dict = json.loads(json_str)
        tool = ToolClz(**argument_dict)
        return tool
    except (json.JSONDecodeError, pydantic.ValidationError) as e:
        if depth >= max_depth:
            raise Exception("Invalid Tool JSON. Max depth reached.")

        msg = f"""# Instructions
        You must fix the invalid JSON provided to match the tool.
        Use the error message below to fix the JSON.

        # JSON
        {json_str}

        # Error
        {str(e)}
        """
        tools, tool_choice = generate_tool_json(
            ToolClz,
            force_text_response=False,
            required_tool=ToolClz
        )
        response = litellm_request(
            session.model.ssl_verify,
            session.model.name,
            session.model.max_retries,
            session.model.api_key,
            generate_message_json(LLMUserMessage(content=msg)),
            tools.openapi_json(),
            tool_choice,
            verbose=session.verbose,
            headers=session.model.headers,
            organization=session.model.organization,
            api_base=session.model.api_base
        )
        return sync_tool_from_json_str(
            session,
            response.choices[0].message.tool_calls[0].function.arguments,
            ToolClz,
            depth + 1,
            max_depth
        )



from .version import VERSION, version_short

from .prompt import TypedPrompt

from .models import (
    LLMRequest,
    LLMSession,
    LLMModel,
    LLMResponse,
    LLMMessage,
    LLMUserMessage,
    LLMAssistantMessage,
    LLMToolResultMessage
)

from .client import request, async_request, typed_request, async_typed_request

from .tool import create_tool_from_function, Tool, ToolCollection

__all__ = [
    'VERSION', 'version_short',
    'LLMRequest', 'LLMSession', 'LLMModel', 'LLMResponse',
    'LLMMessage', 'LLMUserMessage', 'LLMAssistantMessage', 'LLMToolResultMessage',
    'request', 'async_request', 'typed_request', 'async_typed_request',
    'create_tool_from_function', 'Tool', 'ToolCollection',
    'TypedPrompt'
]



import json
from typing import List, Optional, Dict, Union, Any, Type
from pydantic import BaseModel, Field
from .tool import Tool


class LLMMessage(BaseModel):

    def openai_json(self):
        raise NotImplementedError("This method must be implemented by the subclass.")


class LLMSystemMessage(LLMMessage):
    content: str

    def openai_json(self):
        return {
            "role": "system",
            "content": self.content
        }


class ImageURL(BaseModel):
    url: str = Field(description="The URL of the image. Can be a web URL or a Base64 encoded image.")


class LLMUserMessage(LLMMessage):
    content: Union[str, List[Union[str, ImageURL]]] = Field(
        description="The content of the message. Either a list of strings or Image URLs."
    )

    def openai_json(self):
        content = None
        if isinstance(self.content, str):
            content = self.content
        else:
            content = []
            for item in self.content:
                if isinstance(item, str):
                    content.append({
                        "type": "text",
                        "text": item
                    })
                else:
                    content.append({
                        "type": "image_url",
                        "image_url": {
                            "url": item.url
                        }
                    })
        return {
            "role": "user",
            "content": content
        }


class LLMAssistantToolCall(LLMMessage):
    id: str = Field(description="The id of the tool call")
    tool: Tool = Field(description="The tool used in the call")
    args: str = Field(description="The arguments used in the tool call")

    def openai_json(self):
        return {
            "id": self.id,
            "function": {
                "name": self.tool.name,
                "arguments": self.args
            }
        }


class LLMAssistantMessage(LLMMessage):
    content: str = Field(
        description="The content of the message. Either a list of strings or Image URLs."
    )

    def openai_json(self):
        return {
            "role": "assistant",
            "content": self.content
        }


class LLMToolResultMessage(LLMMessage):
    id: str = Field(description="The id of the tool call")
    content: Any = Field(description="The result of the tool call")

    def openai_json(self):
        return {
            "role": "tool",
            "content": self.content,
            "tool_call_id": self.id
        }


class LLMModel(BaseModel):
    name: str = Field(
        description="The name of the model using the naming conventions of LiteLLM. "
                    "More info here: https://docs.litellm.ai/docs/providers"
    )
    api_key: str = Field(description="The API key for the model provider.")
    organization: Optional[str] = Field(
        description="The organization ID for the model provider.",
        default=None
    )
    api_base: Optional[str] = Field(
        description="If using a Proxy or 3rd Party vendor, this is the base url for requests.",
        default=None
    )
    headers: Optional[Dict[str, str]] = Field(
        description="Additional headers to include with the requests. "
                    "More info here: https://docs.litellm.ai/docs/providers/openai#using-helicone-proxy-with-litellm",
        default=None
    )
    max_retries: int = Field(
        default=3,
        description="The maximum number of retries for a step in the pipeline."
    )
    ssl_verify: bool = Field(
        description="Whether to verify the SSL certificate of the model provider.",
        default=True
    )


class LLMSession(BaseModel):
    model: Optional[LLMModel] = Field(description="The model used in the session", default=None)
    messages: List[LLMMessage] = Field(description="The messages in the session", default=[])
    tools: List[Type[Tool]] = Field(description="The tools available in the session", default=[])
    verbose: bool = Field(description="Whether to log verbose output", default=False)

    def merge(self, other: 'LLMSession') -> 'LLMSession':
        return LLMSession(
            model=self.model or other.model,
            messages=self.messages + other.messages,
            tools=self.tools + other.tools,
            verbose=self.verbose or other.verbose
        )


class LLMRequest(BaseModel):
    message: Optional[LLMUserMessage] = Field(description="The prompt to use for the request", default=None)
    tool_results: List[LLMToolResultMessage] = Field(description="Tool results for this request", default=[])
    tools: List[Type[Tool]] = Field(
        description="The tools to use for the request in addition to the session tools.",
        default=[]
    )
    required_tool: Optional[Type[Tool]] = Field(
        description="The tool that must be used for the request",
        default=None
    )
    force_text_response: bool = Field(
        description="Whether to force the response to be text.",
        default=False
    )

    def merge(self, other: 'LLMRequest') -> 'LLMRequest':
        return LLMRequest(
            message=self.message or other.message,
            tool_results=self.tool_results + other.tool_results,
            tools=self.tools + other.tools,
            required_tool=self.required_tool or other.required_tool,
            force_text_response=self.force_text_response or other.force_text_response
        )


class LLMResponse(BaseModel):
    message: Optional[LLMAssistantMessage] = Field(description="The message from the assistant", default=None)
    tool_calls: List[LLMAssistantToolCall] = Field(description="The tools to use for the request", default=[])
    raw: BaseModel = Field(description="The raw response from the model provider")
